@@ -111,6 +111,12 @@ static DOMString quoteStringIfNeeded(const DOMString &string)
     return isCSSTokenizerIdentifier(string) ? string : quoteString(string);
 }
 
+static inline bool isInitialOrInherit(const CSSValueImpl *value)
+{
+    const unsigned short t = value->cssValueType();
+    return (t == CSSValue::CSS_INHERIT || t == CSSValue::CSS_INITIAL);
+}
+
 CSSStyleDeclarationImpl::CSSStyleDeclarationImpl(CSSRuleImpl *parent)
     : StyleBaseImpl(parent)
 {
@@ -214,10 +220,11 @@ DOMString CSSStyleDeclarationImpl::getPropertyValue( int propertyID ) const
     }
     case CSS_PROP_BACKGROUND:
     {
-        const int properties[6] = { CSS_PROP_BACKGROUND_IMAGE, CSS_PROP_BACKGROUND_REPEAT,
-                                    CSS_PROP_BACKGROUND_ATTACHMENT,CSS_PROP_BACKGROUND_POSITION_X,
-                                    CSS_PROP_BACKGROUND_POSITION_Y, CSS_PROP_BACKGROUND_COLOR };
-        return getLayeredShortHandValue( properties, 6 );
+        // 'clip' must come after 'origin' in this array
+        const int properties[9] = { CSS_PROP_BACKGROUND_IMAGE, CSS_PROP_BACKGROUND_REPEAT, CSS_PROP_BACKGROUND_ATTACHMENT,
+                                    CSS_PROP_BACKGROUND_POSITION_X, CSS_PROP_BACKGROUND_POSITION_Y, CSS_PROP_BACKGROUND_SIZE,
+                                    CSS_PROP_BACKGROUND_ORIGIN, CSS_PROP_BACKGROUND_CLIP, CSS_PROP_BACKGROUND_COLOR };
+        return getLayeredShortHandValue( properties, 9 );
     }
     case CSS_PROP_BORDER:
     {
@@ -338,18 +345,62 @@ DOMString CSSStyleDeclarationImpl::getCommonValue(const int* properties, int num
 
 DOMString CSSStyleDeclarationImpl::get4Values( const int* properties ) const
 {
-    DOMString res;
-    for ( int i = 0 ; i < 4 ; ++i ) {
-        if (!isPropertyImplicit(properties[i])) {
-            CSSValueImpl* value = getPropertyCSSValue( properties[i] );
-            if ( !value ) { // apparently all 4 properties must be specified.
-                return DOMString();
-            }
-            if ( i > 0 )
-                res += " ";
-            res += value->cssText();
+    // Assume the properties are in the order top, right, bottom, left.
+    QVector<DOMString> values(4);
+    for (int i = 0 ; i < 4 ; ++i) {
+        CSSValueImpl *val = getPropertyCSSValue(properties[i]);
+        // All 4 properties must be specified.
+        if (!val || isInitialOrInherit(val)) {
+            return DOMString();
+        } else {
+            values[i] = val->cssText();
         }
     }
+
+    // Reduce shorthand.
+    if (values.at(1) == values.at(3)) { // right/left
+        values.remove(3);
+        if (values.at(0) == values.at(2)) { // top/bottom
+            values.remove(2);
+            if (values.at(0) == values.at(1)) {
+                values.remove(1);
+            }
+        }
+    }
+
+    DOMString res;
+    const int valuesSize = values.size();
+    for (int i = 0; i < valuesSize; ++i) {
+        if (!res.isNull()) {
+            res += " ";
+        }
+        res += values.at(i);
+    }
+
+    return res;
+}
+
+static inline DOMString posXYSize_string_helper(DOMString &bPosX, DOMString &bPosY, DOMString &bSize) {
+    DOMString res, position;
+
+    if (!bPosX.isEmpty() && !bPosY.isEmpty()) {
+        position = bPosX + DOMString(" ") + bPosY;
+    } else if (bPosX.isEmpty() && !bPosY.isEmpty()) {
+        position = DOMString("0% ") + bPosY;
+    } else if (!bPosX.isEmpty() && bPosY.isEmpty()) {
+        position = bPosX + DOMString(" 0%");
+    }
+
+    if (!bSize.isEmpty()) {
+        if (position.isEmpty()) {
+            res = DOMString("0% 0%") + DOMString(" / ") + bSize;
+        } else {
+            res = position + DOMString(" / ") + bSize;
+        }
+    } else {
+        res = position;
+    }
+
     return res;
 }
 
@@ -374,10 +425,11 @@ DOMString CSSStyleDeclarationImpl::getLayeredShortHandValue(const int* propertie
         }
     }
 
-    // Now stitch the properties together.  Implicit initial values are flagged as such and
-    // can safely be omitted.
+    // Now stitch the properties together.
+    // Implicit initial values are flagged as such and can safely be omitted.
     for (i = 0; i < numLayers; i++) {
         DOMString layerRes;
+        DOMString bPosX, bPosY, bSize;
         for (j = 0; j < number; j++) {
             CSSValueImpl* value = 0;
             if (values[j]) {
@@ -396,12 +448,31 @@ DOMString CSSStyleDeclarationImpl::getLayeredShortHandValue(const int* propertie
             }
 
             if (value && !value->isImplicitInitialValue()) {
-                if (!layerRes.isNull())
-                    layerRes += " ";
-                layerRes += value->cssText();
+                // positionX,positionY,size should be handled separately in order
+                // to return a consistent and valid 'background' property string
+                if (properties[j] == CSS_PROP_BACKGROUND_POSITION_X) {
+                    bPosX = value->cssText();
+                } else if (properties[j] == CSS_PROP_BACKGROUND_POSITION_Y) {
+                    bPosY = value->cssText();
+                } else if (properties[j] == CSS_PROP_BACKGROUND_SIZE) {
+                    bSize = value->cssText();
+                } else {
+                    if (!layerRes.isNull())
+                        layerRes += " ";
+                    layerRes += value->cssText();
+                }
             }
         }
 
+        // now add positionX,positionY,size
+        DOMString posXYSize = posXYSize_string_helper(bPosX, bPosY, bSize);
+        if (!posXYSize.isEmpty()) {
+            if (!layerRes.isNull()) {
+                layerRes += " ";
+            }
+            layerRes += posXYSize;
+        }
+
         if (!layerRes.isNull()) {
             if (!res.isNull())
                 res += ", ";
@@ -586,8 +657,9 @@ static void initShorthandMap(QHash<int, PropertyLonghand>& shorthandMap)
         CSS_PROP_BACKGROUND_POSITION_X,
         CSS_PROP_BACKGROUND_POSITION_Y,
         CSS_PROP_BACKGROUND_REPEAT,
-        CSS_PROP__KHTML_BACKGROUND_SIZE,
-        CSS_PROP_BACKGROUND_SIZE
+        CSS_PROP_BACKGROUND_SIZE,
+        CSS_PROP_BACKGROUND_ORIGIN,
+        CSS_PROP_BACKGROUND_CLIP
     };
     SET_SHORTHAND_MAP_ENTRY(shorthandMap, CSS_PROP_BACKGROUND, backgroundProperties);
 
@@ -613,13 +685,6 @@ static void initShorthandMap(QHash<int, PropertyLonghand>& shorthandMap)
         CSS_PROP_BORDER_BOTTOM_RIGHT_RADIUS
     };
     SET_SHORTHAND_MAP_ENTRY(shorthandMap, CSS_PROP_BORDER_RADIUS, borderRadiusProperties);
-    static const int prefixedBorderRadiusProperties[] = {
-        CSS_PROP__KHTML_BORDER_TOP_RIGHT_RADIUS,
-        CSS_PROP__KHTML_BORDER_TOP_LEFT_RADIUS,
-        CSS_PROP__KHTML_BORDER_BOTTOM_LEFT_RADIUS,
-        CSS_PROP__KHTML_BORDER_BOTTOM_RIGHT_RADIUS
-    };
-    SET_SHORTHAND_MAP_ENTRY(shorthandMap, CSS_PROP__KHTML_BORDER_RADIUS, prefixedBorderRadiusProperties);
 
     static const int markerProperties[] = {
         CSS_PROP_MARKER_START,
